/*
   Source for x86 emulator
   Copyright (c) 2003, Chris Eagle
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option) 
   any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
   more details.
   
   You should have received a copy of the GNU General Public License along with 
   this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
   Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <malloc.h>

#include "x86defs.h"

//masks to clear out bytes appropriate to the sizes above
dword SIZE_MASKS[] = {0, 0x000000FF, 0x0000FFFF, 0, 0xFFFFFFFF};

//masks to clear out bytes appropriate to the sizes above
dword SIGN_BITS[] = {0, 0x00000080, 0x00008000, 0, 0x80000000};

//masks to clear out bytes appropriate to the sizes above
qword CARRY_BITS[] = {0, 0x00000100, 0x00010000, 0, 0x100000000};

byte BITS[] = {0, 8, 16, 0, 32};

const uchar parityValues[256] = {
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
   1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
   };

//The cpu
regs cpu = {{0, 0, 0, 0, 0, 0, 0, 0}, 0, 2};

static ProgramData *theProgram;
static uint stackBottom;

void setProgram(ProgramData *pd) {
   theProgram = pd;
   initStack(0x1000);
   esp = pd->stackTop;
   cpu.eip = pd->entry;
}

void initStack(unsigned int size) {
   theProgram->stack = (unsigned char *) realloc(theProgram->stack, size);
   theProgram->stackSize = size;
   stackBottom = theProgram->stackTop - size;
}

void growStack(unsigned int more) {
   theProgram->stack = (unsigned char *) realloc(theProgram->stack, theProgram->stackSize + more);
   theProgram->stackSize += more;
   stackBottom = theProgram->stackTop - theProgram->stackSize;
}

void shrinkStack(unsigned int less) {
   theProgram->stack = (unsigned char *) realloc(theProgram->stack, theProgram->stackSize - less);
   theProgram->stackSize -= less;
   stackBottom = theProgram->stackTop - theProgram->stackSize;
}

//sign extension functions
//byte->word
word sebw(byte val) {
   short result = (char)val;
   return (word) result;
}

//word->dword
dword sewd(word val) {
   int result = (short)val;
   return (dword) result;
}

//byte->dword
dword sebd(byte val) {
   int result = (char)val;
   return (dword) result;
}

//dword->qword
qword sedq(dword val) {
   quad result = (int)val;
   return (qword) result;
}

//return a byte
byte readByte(dword addr) {
   if (addr >= stackBottom && addr < theProgram->stackTop) {
      dword internal = theProgram->stackTop - addr;
      return theProgram->stack[internal - 1];
   }
#ifdef __IDP__
   //interface to IDA to read a byte
   //from virtual program space
   return get_byte(addr);
#else
   //assume user provided program space
   return theProgram->program[addr - theProgram->org];
#endif
}

DirtyList *dirtyHead = NULL;

void tagAsDirty(dword addr) {
   addr &= 0xFFFFFFF0;
   DirtyList *prev = NULL, *temp = dirtyHead;
   if (dirtyHead) {
      while (temp && temp->addr < addr) {
         prev = temp;
         temp = temp->next;
      }
      if (temp && temp->addr == addr) return;
      if (prev) {
         prev->next = new DirtyList(addr, temp);
         return;
      }
   }
   dirtyHead = new DirtyList(addr, temp);
}

void freeDirtyList() {
   DirtyList *prev, *temp = dirtyHead;
   while (temp) {
      prev = temp;
      temp = temp->next;
      delete prev;
   }
   dirtyHead = NULL;
}

//don't interface to IDA's get_word/long routines so
//that we can detect stack usage in readByte
word readWord(dword addr) {
   word result = readByte(addr + 1);
   result <<= 8;
   return result | readByte(addr);
}

dword readDword(dword addr) {
   dword result = readWord(addr + 2);
   result <<= 16;
   return result | readWord(addr);
}

//all reads from memory should be through this function
dword readMem(dword addr, byte size) {
   int result = 0;
   switch (size) {
      case SIZE_BYTE:
         result = (int) readByte(addr);
         break;
      case SIZE_WORD:
         result = (int) readWord(addr);
         break;
      case SIZE_DWORD:
         result = (int) readDword(addr);
         break;
   }
   return result;
}

//store a byte
void writeByte(dword addr, byte val) {
   if (addr >= stackBottom) { //on the internal stack
      dword internal = theProgram->stackTop - addr;
      theProgram->stack[internal - 1] = val;
      tagAsDirty(addr);
   }
   else {
#ifdef __IDP__
   //interface to IDA to write a byte
   //to virtual program space
//   put_byte(addr, val);
   patch_byte(addr, val);
#else
   //no IDA so assume user supplied program space
   theProgram->program[addr - theProgram->org] = val;
#endif
   }
}

//don't interface to IDA's put_word/long routines so
//that we can detect stack usage in writeByte
void writeWord(dword addr, word val) {
   writeByte(addr, (byte)val);
   writeByte(addr + 1, (byte)(val >> 8));
}

void writeDword(dword addr, dword val) {
   writeWord(addr, (word)val);
   writeWord(addr + 2, (word)(val >> 16));
}

//all writes to memory should be through this function
void writeMem(dword addr, dword val, byte size) {
   switch (size) {
      case SIZE_BYTE:
         writeByte(addr, (byte)val);
         break;
      case SIZE_WORD:
         writeWord(addr, (word)val);
         break;
      case SIZE_DWORD:
         writeDword(addr, val);
         break;
   }
}

void push(dword val, byte size) {
   esp -= size;
   writeMem(esp, val, size);
}

dword pop(byte size) {
   dword result = readMem(esp, size);
   esp += size;
   return result;
}

//read according to specified n from eip location 
dword fetch(byte n) {
   dword result = readMem(cpu.eip, n);
   cpu.eip += n;
   return result;
}

//fetch an unsigned quantity
dword fetchu(byte n) {
   return fetch(n) & SIZE_MASKS[n];
}

void ExGx16(inst *i) {
   byte modrm = fetchu(SIZE_BYTE);
   byte mod = MOD(modrm);
   byte rm = RM(modrm);
   dword disp = 0;
   if (mod != MOD_3) {
      switch (rm) {
         case 0:
            i->dest = ebx + esi;
            break;
         case 1:
            i->dest = ebp + esi;
            break;
         case 2:
            i->dest = ebx + edi;
            break;
         case 3:
            i->dest = ebp + edi;
            break;
         case 4:
            i->dest = esi;
            break;
         case 5:
            i->dest = edi;
            break;
         case 6:
            i->dest = ebp;
            break;
         case 7:
            i->dest = ebx;
            break;
      }
   }
   i->destType = mod == MOD_3 ? TYPE_REG : TYPE_MEM;
   switch (mod) {
      case MOD_0:
         if (rm == 6) {
            i->dest = fetch(SIZE_WORD);
         }
         break;
      case MOD_1:
         disp = (char) fetch(SIZE_BYTE);
         break;
      case MOD_2:
         disp = (int) fetch(SIZE_WORD);
         break;
      case MOD_3:
         i->dest = rm;
         break;
   }
   if (i->destType == TYPE_MEM) {
      i->dest += disp;
      i->dest &= SIZE_MASKS[SIZE_WORD];
   }
   i->source = REG(modrm);
   i->srcType = TYPE_REG;
}

//i - the current instruction
void ExGx(inst *i) {
   if (i->prefix & PREFIX_ADDR) {
      ExGx16(i);
      return;
   }
   byte modrm = fetchu(SIZE_BYTE);
   byte mod = MOD(modrm);
   byte rm = RM(modrm);
   byte sib;
   dword disp = 0;
   byte hasSib = 0;
   if (mod != MOD_3) {
      if (rm == 4) {
         sib = fetchu(SIZE_BYTE);
         hasSib = 1;
      }
      else {
         i->dest = cpu.general[rm];
      }
   }
   i->destType = mod == MOD_3 ? TYPE_REG : TYPE_MEM;
   switch (mod) {
      case MOD_0:
         if (rm == 5) {
            i->dest = fetch(SIZE_DWORD);
         }
         break;
      case MOD_1:
         disp = (char) fetch(SIZE_BYTE);
         break;
      case MOD_2:
         disp = (int) fetch(SIZE_DWORD);
         break;
      case MOD_3:
         i->dest = rm;
         break;
   }
   if (i->destType == TYPE_MEM) {
      i->dest += disp;
      if (hasSib) {
         dword index = INDEX(sib);
         index = index == 4 ? 0 : cpu.general[index] * SCALE(sib);
         i->dest += index;
         dword base = BASE(sib);
         if (base == 5 && mod == MOD_0) {
            i->dest += fetch(SIZE_DWORD);
         }
         else {
            i->dest += cpu.general[base];
         }
      }
   }
   i->source = REG(modrm);
   i->srcType = TYPE_REG;
}

void GxEx16(inst *i) {
   byte modrm = fetchu(SIZE_BYTE);
   byte mod = MOD(modrm);
   byte rm = RM(modrm);
   dword disp = 0;
   if (mod != MOD_3) {
      switch (rm) {
         case 0:
            i->source = ebx + esi;
            break;
         case 1:
            i->source = ebp + esi;
            break;
         case 2:
            i->source = ebx + edi;
            break;
         case 3:
            i->source = ebp + edi;
            break;
         case 4:
            i->source = esi;
            break;
         case 5:
            i->source = edi;
            break;
         case 6:
            i->source = ebp;
            break;
         case 7:
            i->source = ebx;
            break;
      }
   }
   i->srcType = mod == MOD_3 ? TYPE_REG : TYPE_MEM;
   switch (mod) {
      case MOD_0:
         if (rm == 6) {
            i->source = fetch(SIZE_WORD);
         }
         break;
      case MOD_1:
         disp = (char) fetch(SIZE_BYTE);
         break;
      case MOD_2:
         disp = (int) fetch(SIZE_WORD);
         break;
      case MOD_3:
         i->source = rm;
         break;
   }
   if (i->srcType == TYPE_MEM) {
      i->source += disp;
      i->source &= SIZE_MASKS[SIZE_WORD];
   }
   i->dest = REG(modrm);
   i->destType = TYPE_REG;
}

//i - the current instruction
void GxEx(inst *i) {
   if (i->prefix & PREFIX_ADDR) {
      GxEx16(i);
      return;
   }
   byte modrm = fetchu(SIZE_BYTE);
   byte mod = MOD(modrm);
   byte rm = RM(modrm);
   byte sib;
   dword disp = 0;
   byte hasSib = 0;
   if (mod != MOD_3) {
      if (rm == 4) {
         sib = fetchu(SIZE_BYTE);
         hasSib = 1;
      }
      else {
         i->source = cpu.general[rm];
      }
   }
   i->srcType = mod == MOD_3 ? TYPE_REG : TYPE_MEM;
   switch (mod) {
      case MOD_0:
         if (rm == 5) {
            i->source = fetch(SIZE_DWORD);
         }
         break;
      case MOD_1:
         disp = (char) fetch(SIZE_BYTE);
         break;
      case MOD_2:
         disp = (int) fetch(SIZE_DWORD);
         break;
      case MOD_3:
         i->source = rm;
         break;
   }
   if (i->srcType == TYPE_MEM) {
      i->source += disp;
      if (hasSib) {
         dword index = INDEX(sib);
         index = index == 4 ? 0 : cpu.general[index] * SCALE(sib);
         i->source += index;
         dword base = BASE(sib);
         if (base == 5 && mod == MOD_0) {
            i->source += fetch(SIZE_DWORD);
         }
         else {
            i->source += cpu.general[base];
         }
      }
   }
   i->dest = REG(modrm);
   i->destType = TYPE_REG;
}

void A_Ix(inst *i) {
   i->dest = 0;
   i->destType = TYPE_REG;
   i->source = fetch(i->opsize);
   i->srcType = TYPE_IMM;
}

void setStandardModes(inst *i) {
   i->opsize = i->opcode & 1 ? i->opsize : SIZE_BYTE;
   switch (i->opcode & 0x7) {
      case 0: case 1:
         ExGx(i);
         break;
      case 2: case 3:
         GxEx(i);
         break;
      case 4: case 5:
         A_Ix(i);
         break;
   }
}

dword getSourceOperand(inst *i) {
   dword mask = SIZE_MASKS[i->opsize];
   switch (i->srcType) {
      case TYPE_REG:
         if (i->opsize == SIZE_BYTE && i->source >= 4) {
            //AH, CH, DH, BH
            return (cpu.general[i->source - 4] >> 8) & mask;
         }
         return cpu.general[i->source] & mask;
      case TYPE_IMM:
         return i->source & mask;
      case TYPE_MEM:
         return readMem(i->source, i->opsize) & mask;
   }
   return 0;
}

dword getDestOperand(inst *i) {
   dword mask = SIZE_MASKS[i->opsize];
   if (i->destType == TYPE_REG) {
      if (i->opsize == SIZE_BYTE && i->dest >= 4) {
         //AH, CH, DH, BH
         return (cpu.general[i->dest - 4] >> 8) & mask;
      }
      return cpu.general[i->dest] & mask;
   }
   else {
      return readMem(i->dest, i->opsize) & mask;
   }
}

void storeDest(inst *i, dword val) {
   dword mask = SIZE_MASKS[i->opsize];
   val &= mask;
   if (i->destType == TYPE_REG) {
      if (i->opsize == SIZE_BYTE && i->dest >= 4) {
         //AH, CH, DH, BH
         cpu.general[i->dest - 4] &= ~H_MASK;
         cpu.general[i->dest - 4] |= (val << 8); 
      }
      else {
         cpu.general[i->dest] &= ~SIZE_MASKS[i->opsize];
         cpu.general[i->dest] |= val; 
      }
   }
   else {
      writeMem(i->dest, val, i->opsize);
   }
}

//deal with sign, zero, and parity flags
void setEflags(qword val, byte size) {
   val &= SIZE_MASKS[size]; //mask off upper bytes
   if (val) CLEAR(ZF);
   else SET(ZF);
   if (val & SIGN_BITS[size]) SET(SF);
   else CLEAR(SF);
   if (parityValues[val & 0xFF]) SET(PF);
   else CLEAR(PF);
}

void checkAddOverflow(dword op1, dword op2, dword sum, byte size) {
   dword mask = SIGN_BITS[size];
   if ((op1 & op2 & ~sum & mask) || (~op1 & ~op2 & sum & mask)) SET(OF);
   else CLEAR(OF);
}

dword add(qword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   qword result = (op1 & mask) + (op2 & mask);
   if (result & CARRY_BITS[size]) SET(CF);
   else CLEAR(CF);
   checkAddOverflow((dword)op1, op2, (dword)result, size);
   setEflags(result, size);
   return (dword) result & mask;
}

dword adc(qword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   qword result = (op1 & mask) + (op2 & mask) + C;
   if (result & CARRY_BITS[size]) SET(CF);
   else CLEAR(CF);
   checkAddOverflow((dword)op1, op2, (dword)result, size);
   setEflags(result, size);
   return (dword) result & mask;
}

void checkSubOverflow(dword op1, dword op2, dword diff, byte size) {
   dword mask = SIGN_BITS[size];
   if ((op1 & ~op2 & ~diff & mask) || (~op1 & op2 & diff & mask)) SET(OF);
   else CLEAR(OF);
}

dword sub(qword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   qword result = (op1 & mask) - (op2 & mask);
   if (result & CARRY_BITS[size]) SET(CF);
   else CLEAR(CF);
   checkSubOverflow((dword)op1, op2, (dword)result, size);
   setEflags(result, size);
   return (dword) result & mask;
}

dword sbb(qword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   qword result = (op1 & mask) + (op2 & mask) - C;
   if (result & CARRY_BITS[size]) SET(CF);
   else CLEAR(CF);
   checkSubOverflow((dword)op1, op2, (dword)result, size);
   setEflags(result, size);
   return (dword) result & mask;
}

dword AND(dword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   dword result = (op1 & mask) & (op2 & mask);
   CLEAR(CF | OF);
   setEflags(result, size);
   return result & mask;
}

dword OR(dword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   dword result = (op1 & mask) | (op2 & mask);
   CLEAR(CF | OF);
   setEflags(result, size);
   return result & mask;
}

dword XOR(dword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   dword result = (op1 & mask) ^ (op2 & mask);
   CLEAR(CF | OF);
   setEflags(result, size);
   return result & mask;
}

void cmp(qword op1, dword op2, byte size) {
   dword mask = SIZE_MASKS[size];
   qword result = (op1 & mask) - (op2 & mask);
   if (result & CARRY_BITS[size]) SET(CF);
   else CLEAR(CF);
   checkSubOverflow((dword)op1, op2, (dword)result, size);
   setEflags(result, size);
}

dword inc(qword op1, byte size) {
   dword oldCarry = C;
   op1 = add(op1, 1, size);
   CLEAR(CF);
   cpu.eflags |= oldCarry;
   return (dword) op1;
}

dword dec(qword op1, byte size) {
   dword oldCarry = C;
   op1 = sub(op1, 1, size);
   CLEAR(CF);
   cpu.eflags |= oldCarry;
   return (dword) op1;
}

void checkLeftOverflow(dword result, byte size) {
   dword msb = result & SIGN_BITS[size];
   if ((msb && C) || (!msb && NC)) CLEAR(OF);
   else SET(OF);
}

dword rol(qword op, byte amt, byte size) {
   if (amt) {
      op = op & SIZE_MASKS[size];
      op = (op >> (BITS[size] - amt)) | (op << amt);
      if (op & 1) SET(CF);
      else CLEAR(CF);
      if (amt == 1) {
         checkLeftOverflow((dword)op, size);
      }
   }
   return (dword) op & SIZE_MASKS[size];
}
                               
dword ror(qword op, byte amt, byte size) {
   if (amt) {
      op = op & SIZE_MASKS[size];
      op = (op << (BITS[size] - amt)) | (op >> amt);
      if (op & SIGN_BITS[size]) SET(CF);
      else CLEAR(CF);
      if (amt == 1) {
         dword shift = (dword)op << 1;
         shift = (shift ^ (dword)op) & SIGN_BITS[size];
         if (shift) SET(OF);
         else CLEAR(OF);
      }
   }
   return (dword) op & SIZE_MASKS[size];
}
                               
dword rcl(qword op, byte amt, byte size) {
   if (amt) {
      dword carryPos = 1 << (BITS[size] - amt);
      dword carry = carryPos & (dword)op;
      op = op & SIZE_MASKS[size];
      op = (op >> (BITS[size] - amt + 1)) | (op << amt);
      if (C) op |= carryPos;
      else op &= ~carryPos;
      if (carry) SET(CF);
      else CLEAR(CF);
      if (amt == 1) {
         checkLeftOverflow((dword)op, size);
      }
   }
   return (dword) op & SIZE_MASKS[size];
}
                               
dword rcr(qword op, byte amt, byte size) {
   if (amt) {
      dword carryPos = 1 << (amt - 1);
      dword carry = (1 << (amt - 1)) & (dword)op;
      op = op & SIZE_MASKS[size];
      op = (op << (BITS[size] - amt + 1)) | (op >> amt);
      if (C) op |= carryPos;
      else op &= ~carryPos;
      if (op & carryPos) SET(CF);
      else CLEAR(CF);
      if (amt == 1) {
         dword shift = (dword)op << 1;
         shift = (shift ^ (dword)op) & SIGN_BITS[size];
         if (shift) SET(OF);
         else CLEAR(OF);
      }
   }
   return (dword) op & SIZE_MASKS[size];
}
                             
dword shl(qword op, byte amt, byte size) {
   if (amt) {
      if (amt == 1) {
         checkLeftOverflow((dword)op, size);
      }
      op <<= amt;
      if (op & CARRY_BITS[size]) SET(CF);
      else CLEAR(CF);
   }
   setEflags(op, size);
   return (dword) op & SIZE_MASKS[size];
}

//mask op down to size before calling
dword shiftRight(qword op, byte amt, byte size) {
   if (amt) {
      dword final_carry = 1 << (amt - 1);
      if (op & final_carry) SET(CF);
      else CLEAR(CF);
      op >>= amt;
   }
   setEflags(op, size);
   return (dword) op;
}
                               
dword shr(qword op, byte amt, byte size) {
   if (amt == 1) {
      if (op & SIGN_BITS[size]) SET(OF);
      else CLEAR(OF);
   }
   return shiftRight(op & SIZE_MASKS[size], amt, size);
}
                               
dword sar(qword op, byte amt, byte size) {
   op = op & SIZE_MASKS[size];
   switch (size) {
      case SIZE_BYTE:
         op = sedq(sebd((byte)op));
         break;
      case SIZE_WORD:
         op = sedq(sewd((word)op));
         break;
      case SIZE_DWORD:
         op = sedq((dword)op);
         break;
   }
   if (amt == 1) {
      CLEAR(OF);
   }
   return shiftRight(op, amt, size);
}

dword shrd(qword op1, qword bits, byte amt, byte size) {
   if (amt) {
      dword newCarry = 1 << (amt - 1);
      if (op1 & newCarry) SET(CF);
      else CLEAR(CF);
      bits <<= (BITS[size] - amt);
      op1 = ((op1 & SIZE_MASKS[size]) >> amt) | bits;
   }
   setEflags(op1, size);
   return (dword) (op1 & SIZE_MASKS[size]);
}
                               
dword shld(qword op1, qword bits, byte amt, byte size) {
   if (amt) {
      dword newCarry = 1 << (BITS[size] - amt);
      if (op1 & newCarry) SET(CF);
      else CLEAR(CF);
      bits = (bits & SIZE_MASKS[size]) >> (BITS[size] - amt);
      op1 = (op1 << amt) | bits;
   }
   setEflags(op1, size);
   return (dword) (op1 & SIZE_MASKS[size]);
}

void dShift(inst *i) {
   ExGx(i);
   byte amt;
   if ((i->opcode & 7) == 4) {
      amt = fetch(SIZE_BYTE);
   }
   else {
      amt = ecx & 0xFF;
   }
   amt &= 0x1F;
   dword result;
   dword op1 = getDestOperand(i);
   dword op2 = getSourceOperand(i);
   if (i->opcode < 0xA8) {
      result = shld(op1, op2, amt, i->opsize);
   }
   else {
      result = shrd(op1, op2, amt, i->opsize);
   }
   storeDest(i, result);
}
                               
//handle instructions that begin w/ 0x0n
int doZero(inst *i) {
   byte op = i->opcode & 0x0F;
   dword result;
   if ((op & 0x7) < 6) {
      setStandardModes(i);
      dword op1 = getDestOperand(i);
      dword op2 = getSourceOperand(i);
      if (op < 8) { // ADD
         result = add(op1, op2, i->opsize);
      }
      else { // OR
         result = OR(op1, op2, i->opsize);
      }
      storeDest(i, result);
   }
   else {
      switch (op) {
         case 0x06:
//            push(ES, SIZE_WORD);
            break;
         case 0x07:
//            ES = pop(SIZE_WORD);
            break;
         case 0x0C:
//            push(CS, SIZE_WORD);
            break;
      }
   }
   return 1;
}

//handle instructions that begin w/ 0x1n
int doOne(inst *i) {
   byte op = i->opcode & 0x0F;
   dword result;
   if ((op & 0x7) < 6) {
      setStandardModes(i);
      dword op1 = getDestOperand(i);
      dword op2 = getSourceOperand(i);
      if (op < 8) { // ADC
         result = adc(op1, op2, i->opsize);
      }
      else { // SBB
         result = sbb(op1, op2, i->opsize);
      }
      storeDest(i, result);
   }
   else {
      switch (op) {
         case 6:
//            push(SS, SIZE_WORD);
            break;
         case 7:
//            SS = pop(SIZE_WORD);
            break;
         case 0xC:
//            push(DS, SIZE_WORD);
            break;
         case 0xF:
//            DS = pop(SIZE_WORD);
            break;
      }
   }
   return 1;
}

//handle instructions that begin w/ 0x2n
int doTwo(inst *i) {
   byte op = i->opcode & 0x0F;
   dword result;
   if ((op & 0x7) < 6) {
      setStandardModes(i);
      dword op1 = getDestOperand(i);
      dword op2 = getSourceOperand(i);
      if (op < 8) { // AND
         result = AND(op1, op2, i->opsize);
      }
      else { // SUB
         result = sub(op1, op2, i->opsize);
      }
      storeDest(i, result);
   }
   else {
      switch (op) {
         case 0x06:
//            SEG = ES;
            break;
         case 0x07: { //DAA
            dword al = eax & 0xFF;
            if (((al & 0x0F) > 9) || (cpu.eflags & AF)) {
               dword old_al = al;
               SET(AF);
               al += 6;
               if (C || ((al ^ old_al) & 0x10)) SET(CF);
            }
            else {
               CLEAR(AF);
            }
            if (((al & 0xF0) > 0x90) || C) {
               al += 0x60;
               SET(CF);
            }
            else {
               CLEAR(CF);
            }
            eax = (eax & 0xFFFFFF00) | (al & 0xFF);
            setEflags(eax, SIZE_BYTE);
            break;
         }
         case 0x0C:
//            SEG = CS;
            break;
         case 0x0F: { //DAS
            dword al = eax & 0xFF;
            if (((al & 0x0F) > 9) || (cpu.eflags & AF)) {
               dword old_al = al;
               SET(AF);
               al -= 6;
               if (C || ((al ^ old_al) & 0x10)) SET(CF);
            }
            else {
               CLEAR(AF);
            }
            if (((al & 0xFF) > 0x9F) || C) {
               al -= 0x60;
               SET(CF);
            }
            else {
               CLEAR(CF);
            }
            eax = (eax & 0xFFFFFF00) | (al & 0xFF);
            setEflags(eax, SIZE_BYTE);
            break;
         }
      }
   }
   return 1;
}

//handle instructions that begin w/ 0x3n
int doThree(inst *i) {
   byte op = i->opcode & 0x0F;
   if ((op & 0x7) < 6) {
      setStandardModes(i);
      dword op1 = getDestOperand(i);
      dword op2 = getSourceOperand(i);
      if (op < 8) { // XOR
         storeDest(i, XOR(op1, op2, i->opsize));
      }
      else { // CMP
         cmp(op1, op2, i->opsize);
      }
   }
   else {
      switch (op) {
         case 0x06:
//            SEG = SS;
            break;
         case 0x07: {//AAA
            dword al = eax & 0xFF;
            dword ax = eax & 0xFF00;
            if (((al & 0x0F) > 9) || (cpu.eflags & AF)) {
               SET(CF | AF);
               ax += 0x100;
               al += 6;
            }
            else {
               CLEAR(CF | AF);
            }
            ax |= al;
            eax = (eax & 0xFFFF0000) | (ax & 0xFF0F);
            break;
         }
         case 0x0C:
//            SEG = DS;
            break;
         case 0x0F: { //AAS
            dword al = eax & 0xFF;
            dword ax = eax & 0xFF00;
            if (((al & 0x0F) > 9) || (cpu.eflags & AF)) {
               SET(CF | AF);
               ax = (ax - 0x100) & 0xFF00;
               al -= 6;
            }
            else {
               CLEAR(CF | AF);
            }
            ax |= al;
            eax = (eax & 0xFFFF0000) | (ax & 0xFF0F);
            break;
         }
      }
   }
   return 1;
}

//handle instructions that begin w/ 0x4n
int doFour(inst *i) {
   byte op = i->opcode & 0x0F;
   byte reg = op & 7;
   dword mask = SIZE_MASKS[i->opsize];
   //skip source setup, just read the register
   dword result = cpu.general[reg] & mask;
   i->destType = TYPE_REG;
   i->dest = reg;
   if (op < 8) { // INC
      result = inc(result, i->opsize);
   }
   else { // DEC
      result = dec(result, i->opsize);
   }
   storeDest(i, result);
   setEflags(result, i->opsize);
   return 1;
}

//handle instructions that begin w/ 0x5n
int doFive(inst *i) {
   byte op = i->opcode & 0x0F;
   byte reg = op & 7;
   //skip source setup, just setup the destination
   i->destType = TYPE_REG;
   i->dest = reg;
   if (op < 8) { // PUSH
      push(cpu.general[reg], i->opsize);
   }
   else { // POP
      storeDest(i, pop(i->opsize));
   }
   return 1;
}

//handle instructions that begin w/ 0x6n
int doSix(inst *i) {
   byte op = i->opcode & 0x0F;
   dword result;
   int op1, op2;
   //skip source setup, just setup the destination
   i->destType = TYPE_REG;
   switch (op) {
      case 0: //PUSHA/PUSHAD
         result = esp;
         for (i->source = EAX; i->source <= EDI; i->source++) {
            if (i->source != ESP) push(cpu.general[i->source], i->opsize);
            else push(result, i->opsize);
         }
         break;
      case 1: {//POPA/POPAD
         for (int j = EDI; j >= EAX; j--) { //need signed number for this test
            i->dest = (dword)j;
            if (i->dest == ESP) result = pop(i->opsize);
            else storeDest(i, pop(i->opsize));
         }
         i->dest = ESP;
         storeDest(i, result);
         break;
              }
      case 2: //BOUND
         break;
      case 3: //ARPL
         break;
      case 8: //PUSH Iv
         push(fetch(i->opsize), i->opsize);
         break;
      case 9: //IMUL Iv
         GxEx(i);
         op1 = getSourceOperand(i);
         op2 = fetch(i->opsize);  //need to do some size alignment here
         result = op1 * op2;
         storeDest(i, result);
         setEflags(result, i->opsize);
         break;
      case 0xA: //PUSH Ib
         //not certain this should be sign extended
         push(fetch(SIZE_BYTE), i->opsize);
         break;
      case 0xB: //IMUL Ib
         GxEx(i);
         op1 = getSourceOperand(i);
         op2 = fetch(SIZE_BYTE); //need to do some size alignement here
         result = op1 * op2;
         storeDest(i, result);
         setEflags(result, SIZE_BYTE);
         break;
      case 0xC: //INS
         break;
      case 0xD: //INS
         break;
      case 0xE: //OUTS
         break;
      case 0xF: //OUTS
         break;
   }
   return 1;
}

//handle instructions that begin w/ 0x7n
int doSeven(inst *i) {
   byte op = i->opcode & 0x0F;
   dword imm = fetch(i->opsize);
   int branch = 0;
   switch (op) {
      case 0: //JO
         branch = O;
         break;
      case 1: //JNO
         branch = NO;
         break;
      case 2: //B/NAE/C
         branch = B;
         break;
      case 3: //NB/AE/NC
         branch = NB;
         break;
      case 4:  //E/Z
         branch = Z;
         break;
      case 5:  //NE/NZ
         branch = NZ;
         break;
      case 6:  //BE/NA
         branch = BE;
         break;
      case 7:  //NBE/A
         branch = A;
         break;
      case 8: //S
         branch = S;
         break;
      case 9: //NS
         branch = NS;
         break;
      case 0xA: //P/PE
         branch = P;
         break;
      case 0xB: //NP/PO
         branch = NP;
         break;
      case 0xC: //L/NGE
         branch = L;
         break;
      case 0xD: //NL/GE
         branch = GE;
         break;
      case 0xE: //LE/NG
         branch = LE;
         break;
      case 0xF: //NLE/G
         branch = G;
         break;
   }
   if (branch) {
      cpu.eip += (i->opsize == SIZE_BYTE) ? sebd(imm) : imm;
   }
   return 1;
}

//handle instructions that begin w/ 0x8n
int doEight(inst *i) {
   byte op = i->opcode & 0x0F;
   dword op1, op2;
   byte size = op & 1 ? i->opsize : SIZE_BYTE;
   if (op < 4) {  //83 is sign extended byte->dword
                  //is 82 ever actually used?
      byte subop;
      i->opsize = size;
      ExGx(i);  //we will ignore Gx info
      subop = (byte) i->source;
      op2 = fetch((op == 1) ? i->opsize : SIZE_BYTE);
      if (op == 3) op2 = sebd(op2);
      op1 = getDestOperand(i);
      //ADD, OR, ADC, SBB, AND, SUB, XOR, CMP
      switch (subop) {
         case 0: //ADD
            storeDest(i, add(op1, op2, i->opsize));
            break;
         case 1: //OR
            storeDest(i, OR(op1, op2, i->opsize));
            break;
         case 2: //ADC
            storeDest(i, adc(op1, op2, i->opsize));
            break;
         case 3: //SBB
            storeDest(i, sbb(op1, op2, i->opsize));
            break;
         case 4: //AND
            storeDest(i, AND(op1, op2, i->opsize));
            break;
         case 5: //SUB
            storeDest(i, sub(op1, op2, i->opsize));
            break;
         case 6: //XOR
            storeDest(i, XOR(op1, op2, i->opsize));
            break;
         case 7: //CMP
            cmp(op1, op2, i->opsize);
            break;
      }
   }
   else if (op < 8) {
      i->opsize = size;
      ExGx(i);
      if (op < 6) { //TEST
         AND(getSourceOperand(i), getDestOperand(i), i->opsize);
      }
      else { //XCHG
         dword temp = getDestOperand(i);
         storeDest(i, getSourceOperand(i));
         i->dest = i->source;
         i->destType = i->srcType;
         storeDest(i, temp);
      }
   }
   else if (op < 0xC) {  //MOVE
      i->opsize = size;
      setStandardModes(i);
      storeDest(i, getSourceOperand(i));
   }
   else {
      switch (op) {
         case 0xC: //MOVE seg reg - NOT using segment registers at the moment
            i->opsize = SIZE_WORD;
            break;
         case 0xD: //LEA
            GxEx(i); //generate the address
            storeDest(i, i->source); //store the address
            break;
         case 0xE: //MOVE seg reg - NOT using segment registers at the moment
            i->opsize = SIZE_WORD;
            break;
         case 0xF: //POP
            ExGx(i); //no source, just generate destination info
            storeDest(i, pop(i->opsize));
            break;
      }
   }
   return 1;
}

//handle instructions that begin w/ 0x9n
int doNine(inst *i) {
   byte op = i->opcode & 0x0F;
   dword temp;
   i->destType = TYPE_REG;
   if (op < 8) { //0 is actually NOP, but we do XCHG eax, eax here
      i->dest = op & 7;
      temp = cpu.general[i->dest];
      storeDest(i, eax);
      i->dest = 0;
      storeDest(i, temp);
   }
   else {
      switch (op) {
         case 8: //CBW/CWDE
            i->dest = EAX;
            if (i->opsize == SIZE_WORD) storeDest(i, sebw(eax));
            else storeDest(i, sewd(eax));
            break;
         case 9: //CWD/CDQ
            i->dest = EDX;
            temp = eax & SIGN_BITS[i->opsize] ? 0xFFFFFFFF : 0;
            storeDest(i, temp);
            break;
         case 0xA: //CALLF  //not required for my purposes?
            break;
         case 0xB: //FWAIT/WAIT  //not dealing with FP
            break;
         case 0xC: //PUSHF/PUSHFD
            push(cpu.eflags, i->opsize);
            break;
         case 0xD: //POPF/POPFD
            cpu.eflags = pop(i->opsize);
            break;
         case 0xE: //SAHF
            temp = eax >> 8;
            temp &= 0xD5;
            temp |= 2;
            cpu.eflags &= ~SIZE_MASKS[SIZE_BYTE];
            cpu.eflags |= temp;
            break;
         case 0xF: //LAHF
            temp = cpu.eflags & SIZE_MASKS[SIZE_BYTE] << 8;
            eax &= ~H_MASK;
            eax |= temp;
            break;
      }
   }
   return 1;
}

void stepd(byte size) {
   D ? (edi -= size) : (edi += size);
}

void steps(byte size) {
   D ? (esi -= size) : (esi += size);
}

void step(byte size) {
   stepd(size);
   steps(size);
}

//handle instructions that begin w/ 0xAn
int doTen(inst *i) {
   byte op = i->opcode & 0x0F;
   dword data;
   dword rep = i->prefix & PREFIX_REP;
   dword repne = i->prefix & PREFIX_REPNE;
   i->dest = EAX;
   i->destType = TYPE_REG;
   switch (op) {
      case 0: // Segemented MOV moffs
         i->opsize = SIZE_BYTE;
         break;
      case 1: // Segemented MOV moffs
         i->source = fetch(SIZE_DWORD);
         i->srcType = TYPE_MEM;
         storeDest(i, getSourceOperand(i));
         break;
      case 2: // Segemented MOV moffs
         i->opsize = SIZE_BYTE;
         break;
      case 3: // Segemented MOV moffs
         i->dest = fetch(SIZE_DWORD);
         i->destType = TYPE_MEM;
         storeDest(i, eax);
         break;
      case 4:  //MOVS/MOVSB
         i->opsize = SIZE_BYTE;
      case 5:  //MOVS/MOVSW/MOVSD
         do {
            writeMem(edi, readMem(esi, i->opsize), i->opsize);
            step(i->opsize);
         } while (rep && ecx);
         break;
      case 6:  //CMPS/CMPSB
         i->opsize = SIZE_BYTE;
      case 7: //CMPS/CMPSW/CMPSD
         do {
            cmp(readMem(esi, i->opsize), readMem(edi, i->opsize), i->opsize);
            step(i->opsize);
            if (rep && NZ) break;
            if (repne && Z) break;
         } while ((rep | repne) && ecx);
         break;
      case 8: case 9: //TEST
         if (op == 8) {
            i->opsize = SIZE_BYTE;
         }
         data = fetch(i->opsize);
         AND(getDestOperand(i), data, i->opsize);
         break;
      case 0xA: //STOS/STOSB
         i->opsize = SIZE_BYTE;
      case 0xB: //STOS/STOSW/STOSD
         do {
            writeMem(edi, getDestOperand(i), i->opsize);
            stepd(i->opsize);         
         } while (rep && ecx);
         break;
      case 0xC: //LODS/LODSB
         i->opsize = SIZE_BYTE;
      case 0xD: //LODS/LODSW/LODSD
         do {
            storeDest(i, readMem(esi, i->opsize));
            steps(i->opsize);
         } while (rep && ecx);
         break;
      case 0xE: //SCAS/SCASB
         i->opsize = SIZE_BYTE;
      case 0xF: //SCAS/SCASW/SCASD
         do {
            cmp(getDestOperand(i), readMem(edi, i->opsize), i->opsize);
            stepd(i->opsize);
            if (rep && NZ) break;
            if (repne && Z) break;
         } while ((rep | repne) && ecx);
         break;
   }
   return 1;
}

//handle instructions that begin w/ 0xBn
int doEleven(inst *i) {
   byte op = i->opcode & 0x0F;
   i->dest = op & 7;
   i->destType = TYPE_REG;
   if (op < 8) {
      dword data = fetch(SIZE_BYTE);
      if (op < 4) {
         i->opsize = SIZE_BYTE;
         storeDest(i, data);
      }
      else {
         cpu.general[i->dest & 3] &= ~H_MASK;
         data <<= 8;
         cpu.general[i->dest & 3] |= (data & H_MASK);
      }
   }
   else {
      storeDest(i, fetch(i->opsize));
   }
   return 1;
}

//handle instructions that begin w/ 0xCn
int doTwelve(inst *i) {
   byte op = i->opcode & 0x0F;
   byte subop;
   dword delta, temp;
   switch (op) {
      case 0: //
         i->opsize = SIZE_BYTE;
      case 1: // SHFT Group 2
         ExGx(i);
         subop = i->source;
         delta = fetch(SIZE_BYTE) & 0x1F;  //shift amount
         if (delta) {
            temp = getDestOperand(i);
            switch (subop) {
               case 0: //ROL
                  storeDest(i, rol(temp, delta, i->opsize));
                  break;
               case 1: //ROR
                  storeDest(i, ror(temp, delta, i->opsize));
                  break;
               case 2: //RCL
                  storeDest(i, rcl(temp, delta, i->opsize));
                  break;
               case 3: //RCR
                  storeDest(i, rcr(temp, delta, i->opsize));
                  break;
               case 4:  //SHL/SAL
                  storeDest(i, shl(temp, delta, i->opsize));
                  break;
               case 5:  //SHR
                  storeDest(i, shr(temp, delta, i->opsize));
                  break;
               case 7: //SAR
                  storeDest(i, sar(temp, delta, i->opsize));
                  break;
            }
         }
         break;
      case 2: //RETN Iw
         delta = fetchu(SIZE_WORD);
         cpu.eip = pop(SIZE_DWORD);
         esp += delta;
         break;
      case 3: //RETN
         cpu.eip = pop(SIZE_DWORD);
         break;
      case 4:  //LES - NOT using segments now
         break;
      case 5:  //LDS - NOT using segments now
         break;
      case 6:  // MOV
         i->opsize = SIZE_BYTE;
      case 7: // MOV
         ExGx(i);
         storeDest(i, fetch(i->opsize));
         break;
      case 8: //ENTER
         delta = fetchu(SIZE_WORD);
         subop = fetchu(SIZE_BYTE);
         push(ebp, SIZE_DWORD);
         temp = esp;
         if (subop > 0) {
            while (--subop) {
               ebp -= 4;
               push(readMem(ebp, SIZE_DWORD), SIZE_DWORD);
            }
            push(temp, SIZE_DWORD);
         }
         ebp = temp;
         esp -= delta;
         break;
      case 9: //LEAVE
         esp = ebp;
         ebp = pop(SIZE_DWORD);
         break;
      case 0xA: //RETF Iw
         break;
      case 0xB: //RETF
         break;
      case 0xC: case 0xD: case 0xE: //INT 3 = 0xCC, INT Ib, INTO
         if (op == 0xD) subop = fetchu(SIZE_BYTE);  //this is the interrupt vector
         else subop = op == 0xC ? 3 : 4;  //3 == TRAP, 4 = O
         break;
      case 0xF: //IRET
         break;
   }
   return 1;
}

//handle instructions that begin w/ 0xDn
int doThirteen(inst *i) {
   byte op = i->opcode & 0x0F;
   byte subop;
   dword delta, temp;
   switch (op) {
      case 0: case 2: //
         i->opsize = SIZE_BYTE;
      case 1: case 3: // SHFT Group 2
         ExGx(i);
         subop = i->source;
         delta = op < 2 ? 1 : ecx & 0x1F;  //shift amount
         temp = getDestOperand(i);
         switch (subop) {
            case 0: //ROL
               storeDest(i, rol(temp, delta, i->opsize));
               break;
            case 1: //ROR
               storeDest(i, ror(temp, delta, i->opsize));
               break;
            case 2: //RCL
               storeDest(i, rcl(temp, delta, i->opsize));
               break;
            case 3: //RCR
               storeDest(i, rcr(temp, delta, i->opsize));
               break;
            case 4:  //SHL/SAL
               storeDest(i, shl(temp, delta, i->opsize));
               break;
            case 5:  //SHR
               storeDest(i, shr(temp, delta, i->opsize));
               break;
            case 7: //SAR
               storeDest(i, sar(temp, delta, i->opsize));
               break;
         }
         break;
      case 4: case 5: {//AAM / AAD
         dword base = fetchu(SIZE_BYTE);
         dword al = eax & 0xFF;
         dword ah = (eax >> 8) & 0xFF;
         dword ax = (op == 4) ? ((al / base) << 8) | (al % base) :
                                (al + ah * base) & 0xFF;
         setEflags(ax, SIZE_WORD);
         eax = (eax & ~SIZE_MASKS[SIZE_WORD]) | ax;
         break;
      }
      case 7: //XLAT/XLATB
         break;
/* Coprocessor escapes         
      case 8: //
         break;
      case 9: //
         break;
      case 0xA: //
         break;
      case 0xB: //
         break;
      case 0xC: //
         break;
      case 0xD: //
         break;
      case 0xE: //
         break;
      case 0xF: //
         break;
*/
   }
   return 1;
}

//handle instructions that begin w/ 0xEn
int doFourteen(inst *i) {
   byte op = i->opcode & 0x0F;
   dword disp;
   dword cond;
   if (op < 4) {
      disp = fetch(SIZE_BYTE);
      if (op < 3) { //LOOPNE/LOOPNZ, LOOPE/LOOPZ, LOOP
         cond = op == 2 ? 1 : op == 0 ? NZ : Z;
         i->dest = ECX;
         i->destType = TYPE_REG;
         storeDest(i, getDestOperand(i) - 1);
         if (getDestOperand(i) && cond) {
            cpu.eip += sebd(disp);
         }
      }
      else {  //JCXZ
         if ((ecx & SIZE_MASKS[i->opsize]) == 0) {
            cpu.eip += sebd(disp);
         }
      }
   }
   switch (op) {
      case 4:  //IN
         break;
      case 5:  //IN
         break;
      case 6:  //OUT
         break;
      case 7: //OUT
         break;
      case 8: //CALL
         disp = fetch(i->opsize);
         if (i->opsize == SIZE_WORD) disp = sewd(disp);
         push(cpu.eip, SIZE_DWORD);
         cpu.eip += disp;
         break;
      case 9: //JMP
         disp = fetch(i->opsize);
         if (i->opsize == SIZE_WORD) disp = sewd(disp);
         cpu.eip += disp;
         break;
      case 0xA: //JMP
         break;
      case 0xB: //JMP
         disp = sebd(fetch(SIZE_BYTE));
         cpu.eip += disp;
         break;
      case 0xC: //IN
         break;
      case 0xD: //IN
         break;
      case 0xE: //OUT
         break;
      case 0xF: //OUT
         break;
   }
   return 1;
}

//handle instructions that begin w/ 0xFn
int doFifteen(inst *i) {
   byte op = i->opcode & 0x0F;
   qword temp, divisor;
   if ((op & 7) > 5) { //subgroup
      byte subop;
      ExGx(i);
      subop = i->source;
      if (op < 8) { //Unary group 3
         if (op == 6) i->opsize = SIZE_BYTE;
         switch (subop) {
            case 0: //TEST
               AND(getDestOperand(i), fetch(i->opsize), i->opsize);
               break;
            case 2: //NOT
               storeDest(i, ~getDestOperand(i));
               break;
            case 3: //NEG
               temp = getDestOperand(i);
               storeDest(i, sub(0, (dword)temp, i->opsize));
               if (temp) SET(CF);
               else CLEAR(CF);
               break;
            case 4: case 5: //MUL: IMUL: (CF/OF incorrect for IMUL
               i->source = i->dest;
               i->srcType = i->destType;
               temp = getSourceOperand(i);
               i->dest = EAX;            //change dest to EAX
               i->destType = TYPE_REG;
               temp *= getDestOperand(i); //multiply by EAX
               if (i->opsize == SIZE_BYTE) {
                  i->opsize = SIZE_WORD;
                  storeDest(i, (dword)temp);
                  temp >>= 8;
               }
               else {
                  storeDest(i, (dword)temp);
                  i->dest = EDX;
                  temp >>= i->opsize == SIZE_WORD ? 16 : 32;
                  storeDest(i, (dword)temp);
               }
               if (temp) SET(CF | OF);
               else CLEAR(CF | OF);
               break;
            case 6: case 7: //DIV: IDIV: (does this work for IDIV?)
               i->source = i->dest;
               i->srcType = i->destType;
               if (i->opsize == SIZE_BYTE) temp = eax & 0xFFFF;
               else if (i->opsize == SIZE_WORD) {
                  temp = ((edx & 0xFFFF) << 16) | (eax & 0xFFFF);
               }
               else {
                  temp = edx;
                  temp <<= 32;
                  temp |= eax;
               }
               divisor = getSourceOperand(i);
               i->dest = EAX;
               i->destType = TYPE_REG;
               storeDest(i, (dword) (temp / divisor));
               i->dest = EDX;
               storeDest(i, (dword) (temp % divisor));
               break;
         }
      }
      else { //group4/5
         dword result;
         if (op == 0xE) i->opsize = SIZE_BYTE; //should only be a group 4
         if (subop < 2) { //INC/DEC
            if (subop == 0) result = inc(getDestOperand(i), i->opsize);
            else result = dec(getDestOperand(i), i->opsize);
            storeDest(i, result);
         }
         else {
            switch (subop) {
               case 2: //CALLN
                  push(cpu.eip, SIZE_DWORD);
                  cpu.eip = getDestOperand(i);
                  break;
               case 3: //CALLF
                  break;
               case 4: //JMPN
                  cpu.eip = getDestOperand(i);
                  break;
               case 5: //JMPF
                  break;
               case 6: //PUSH
                  push(getDestOperand(i), i->opsize);
                  break;
            }
         }
      }
   }
   else {
      switch (op) {
         case 4:  //HLT
            break;
         case 5:  //CMC
            cpu.eflags ^= CF;
            break;
         case 8: //CLC
            CLEAR(CF);
            break;
         case 9: //STC
            SET(CF);
            break;
         case 0xA: //CLI
            CLEAR(IF);
            break;
         case 0xB: //STI
            SET(IF);
            break;
         case 0xC: //CLD
            CLEAR(DF);
            break;
         case 0xD: //STD
            SET(DF);
            break;
      }
   }
   return 1;
}

int doEscape(inst *i) {
   byte opcode = i->opcode = fetchu(SIZE_BYTE);
   dword result;
   switch (opcode & 0xF0) {
      case 0x80: //Jcc
         return doSeven(i); //one byte Jcc handler
         break;
      case 0x90: //SET
         break;
      case 0xA0: //IMUL, SHRD, SHLD
         if (((opcode & 7) == 4) || ((opcode & 7) == 5)) {
            dShift(i);
         }
         break;
      case 0xB0: //MOVZX, MOVSX
         if (opcode == 0xBE) i->opsize = SIZE_BYTE;
         GxEx(i);
         if (opcode > 0xB8) { //MOVSX
            if (opcode == 0xBE) result = sebd((byte)getSourceOperand(i));
            else result = sewd((word)getSourceOperand(i));
         }
         else { //MOVZX
            result = getSourceOperand(i);
         }
         storeDest(i, result);
         break;
   }
   return 1;
}

int executeInstruction() {
   inst currentInst;
   inst *i = &currentInst;
   byte opcode;
   int done = 0;
   i->dest = i->source = i->prefix = 0;
   i->opsize = SIZE_DWORD;  //default
   while (!done) {
      opcode = i->opcode = fetchu(SIZE_BYTE);
      if (opcode == 0x0F) {
         done = doEscape(i);
      }
      else { //one byte opcode
         switch (opcode & 0xF0) {
            case 0x00: //ADD, PUSH, POP, OR 
               done = doZero(i);  //done
               break;
            case 0x10: //ADC, PUSH, POP, SBB
               done = doOne(i);
               break;
            case 0x20: //AND, DAA, SUB, DAS
               switch (opcode) {
                  case 0x26:
                     i->prefix |= PREFIX_ES;
                     break;
                  case 0x2E:
                     i->prefix |= PREFIX_CS;
                     break;
                  default:
                     done = doTwo(i);
                     break;
               }
               break;
            case 0x30: //XOR, AAA, CMP, AAS
               switch (opcode) {
                  case 0x36:
                     i->prefix |= PREFIX_SS;
                     break;
                  case 0x3E:
                     i->prefix |= PREFIX_DS;
                     break;
                  default:
                     done = doThree(i);
                     break;
               }
               break;
            case 0x40: //INC, DEC
               done = doFour(i);
               break;
            case 0x50: //PUSH, POP
               done = doFive(i);
               break;
            case 0x60: //PUSHA, PUSHAD, POPA, POPAD, BOUND, IMUL, INS, INSB, INSW, INSD, OUTS, OUTSB, OUTSW, OUTSD
               switch (opcode) {
                  case 0x64:
                     i->prefix |= PREFIX_FS;
                     break;
                  case 0x65:
                     i->prefix |= PREFIX_GS;
                     break;
                  case 0x66:
                     i->prefix |= PREFIX_SIZE;
                     i->opsize = SIZE_WORD;
                     break;
                  case 0x67:
                     i->prefix |= PREFIX_ADDR;
                     break;
                  default:
                     done = doSix(i);
                     break;
               }
               break;
            case 0x70: //conditional jump
               i->opsize = SIZE_BYTE;
               done = doSeven(i);
               break;
            case 0x80: //IMM1, TEST, XCHG, MOV, LEA, POP
               done = doEight(i);
               break;
            case 0x90: //NOP, XCHG, CBW, CWDE, CWD, CDQ, CALLF, FWAIT, WAIT, PUSHF, PUSHFD, POPF, POPFD, SAHF, LAHF
               done = doNine(i);
               break;
            case 0xA0: //MOV, CMP. TEST, STOS, LODS, SCAS
               done = doTen(i);
               break;
            case 0xB0: //MOV
               done = doEleven(i);
               break;
            case 0xC0: //SHIFT2, RETN, LES, LDS, MOV11, ENTER, LEAVE, RETF, INT, INTO, IRET
               done = doTwelve(i);
               break;
            case 0xD0: //SHIFT2, AAM, AAD, XLAT
               done = doThirteen(i);
               break;
            case 0xE0: //LOOP, Jx, IN, OUT, CALL, JMP
               done = doFourteen(i);
               break;
            case 0xF0: //LOCK, REP, HALT, CMC, UNARY3, CLC, STC, CLI, STI, CLD, STD, INC, DEC
               switch (opcode) {
                  case 0xF0:
                     i->prefix |= PREFIX_LOCK;
                     break;
                  case 0xF2:
                     i->prefix |= PREFIX_REPNE;
                     break;
                  case 0xF3:
                     i->prefix |= PREFIX_REP;
                     break;
                  default:
                     done = doFifteen(i);
                     break;
               }
               break;
         }
      }  
   }
   return 0;
}

/*
int main() {
   inst i;
   executeInstruction(&i);
   fprintf(stderr, "eax = %X\n", eax);
   return 0;
}
*/
