/*
   Kernel module to assist in stripping shiva-0.96 protection
   Copyright (c) 2003, Chris Eagle
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option) 
   any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
   more details.
   
   You should have received a copy of the GNU General Public License along with 
   this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
   Place, Suite 330, Boston, MA 02111-1307 USA
*/

/* shivalkm.c 
 *
 * this LKM can only unwrap binaries that are running
 * you need the password if the binary is password protected
 *
 * Compiling:
 *    gcc -c -I/lib/modules/$(uname -r)/build/include shivalkm.c
 *
 * Using:
 *    load module: insmod shivalkm
 *    unload module: rmmod shivalkm
 *    tail /var/log/messages for info
 *    unwrapped binary will be in ./shivaout
 *
 * tested with a linux 2.4.20 kernel
 */
 
//access_proc_vm is taken from:
/*
 * linux/kernel/ptrace.c
 *
 * (C) Copyright 1999 Linus Torvalds
 *
 * Common interfaces for "ptrace()" which we do not want
 * to continually duplicate across every architecture.
 */

#define MODULE
#define __KERNEL__
#include <linux/kernel.h>
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/highmem.h>
#include <linux/sys.h>
#include <linux/unistd.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <asm/fcntl.h>

MODULE_LICENSE("GPL");

//need this to avoid undefined symbol on insmod
static int errno;

//macros to get ourselves some required syscalls
static inline _syscall0(int,sync);
static inline _syscall1(int,close,int,fd);
static inline _syscall3(int,write,int,fd,void *,buf,off_t,count);
static inline _syscall3(int,open,const char *,file,int,flag,int,mode);

static unsigned char checkme[] = "checkme";

//search for shiva's "checkme" signature
static unsigned int sanityCheck(void *buf, int len) {
   unsigned int checkval = *(unsigned int*)checkme;
   unsigned int index = 0;
   unsigned int *data = (unsigned int*) buf;
   unsigned int end = len / 4;
   while (index < end) {
      if (data[index] == checkval) {
         if (!memcmp(&data[index], checkme, 8)) {
            return index * 4;
         }
      }
      index++;
   }
   return 0;
}

/*
 * Access another process' address space.
 * Source/target buffer must be kernel space, 
 * Do not walk the page table directly, use get_user_pages
 */
 
static int access_proc_vm(struct task_struct *tsk, unsigned long addr, 
                          void *buf, int len)
{
   struct mm_struct *mm;
   struct vm_area_struct *vma;
   struct page *page;
   void *old_buf = buf;
   
   /* Worry about races with exit() */
   task_lock(tsk);
   mm = tsk->mm;
   if (mm)
      atomic_inc(&mm->mm_users);
   task_unlock(tsk);
   if (!mm)
      return 0;
   
   down_read(&mm->mmap_sem);
   /* ignore errors, just check how much was sucessfully transfered */
   while (len) {
      int bytes, ret, offset;
      void *maddr;
      
      ret = get_user_pages(current, mm, addr, 1,
                           0, 1, &page, &vma);
      if (ret <= 0)
         break;
      
      bytes = len;
      offset = addr & (PAGE_SIZE-1);
      if (bytes > PAGE_SIZE-offset)
         bytes = PAGE_SIZE-offset;
      
      flush_cache_page(vma, addr);
      
      maddr = kmap(page);
      memcpy(buf, maddr + offset, bytes);
      flush_page_to_ram(page);
      
      kunmap(page);
      put_page(page);
      len -= bytes;
      buf += bytes;
      addr += bytes;
   }
   up_read(&mm->mmap_sem);
   //mmput(mm);
   
   return (int) (buf - old_buf);
}

void dump_mmap(struct vm_area_struct *mm) {
   if (mm) {
      if ((mm->vm_start & 0xFF000000) == 0x0A000000)
         printk("      (vm_start:%08X) (vm_end:%08X)\n",
      mm->vm_start, mm->vm_end);
      dump_mmap(mm->vm_next);
   }
}

int shivadump_init(void) {
  printk("shivadump loaded...\n");
  struct task_struct *p;
  mm_segment_t old_fs;

  for_each_process(p) {
     if (p->mm) {
        //look for shiva characteristic start address
        if (p->mm->start_code == 0x0A048000) {
           int result;
           int len, fd, v_addr;
           char *buf;
           printk("Task %d may be shiva protected\n", p->pid);
           //all we need to grab are the data pages! That, and to know the
           //start_data, and vm_start
           printk("   (PID:%d) (start data:%08X) (end data:%08X)\n",
                  p->pid, p->mm->start_data, p->mm->end_data);
//           dump_mmap(p->mm->mmap);

           v_addr = p->mm->start_data;
           len = p->mm->end_data - p->mm->start_data;
           buf = (char*) kmalloc(len, GFP_KERNEL);
           result = access_proc_vm(p, p->mm->start_data, buf, len);
           
           printk("vm start addr = 0x%08X\n", p->mm->start_data);
           printk("access_process_vm returned %d\n", result);
           result = sanityCheck(buf, len);
           printk("checkme located at offset 0x%08X\n", result);
           
           old_fs = get_fs();
           set_fs(get_ds());

           //should use sprintf to generate pid specific names
           fd = open("shivaout", O_WRONLY | O_CREAT | O_TRUNC, 0644);
           write(fd, &len, 4);
           write(fd, &v_addr, 4);
           write(fd, buf, len);
           sync();
           close(fd);
           set_fs(old_fs);
           
           kfree(buf);
//           break;
        }
     }
  }
  return 0;
}

void shivadump_exit(void) {
  printk("shivadump unloaded...\n");
}

module_init(shivadump_init);
module_exit(shivadump_exit);

