/*
   Utility to strip shiva-0.96 protection
   Copyright (c) 2003, Chris Eagle
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option) 
   any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
   more details.
   
   You should have received a copy of the GNU General Public License along with 
   this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
   Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <elf.h>
#include <string.h>

#include "x86defs.h"
#include "tea.h"

#define PUSH_r 0
#define CALL 1
#define JUMP 2
#define PUSH_Imm32 3
#define JMP_NearInd 4
#define PUSH_disp_base 5

struct Patch {
   uint type;
   uint loc;
   uint op1;
   uint op2;
   uint len;
};

struct BlockHeader {
   uint magic;
   uint size;
};

struct ProgramBlock {
   uint addr;
   uint filesz;
   uint memsz;
   uchar data[];
};

struct List {
   List *next;
   ProgramBlock *block;
};

struct Block0 {
   uint phtable;
   uint e_entry;
   uint phnum;
   uint phentsize;
   uint ldAddr;
};

#define KEYMAT_SIZE 0x50
#define NUMKEYS 5
#define KEY_DATA_SIZE (KEYMAT_SIZE + (NUMKEYS + 1) * sizeof(int))

uchar checkme[] = "checkme";
uchar magic[] = {0x7F, 'E', 'L', 'F', 0x01, 0x01};

int loadElf(int fd, ProgramData *pd) {
   Elf32_Ehdr elf;
   Elf32_Phdr ph;
   uint org = 0xFFFFFFFF;
   uint maxaddr = 0;
   if (read(fd, &elf, sizeof(Elf32_Ehdr)) != sizeof(Elf32_Ehdr)) {
      close(fd);
      return 1;
   }
   if (memcmp(&elf, magic, sizeof(magic)) || elf.e_type != ET_EXEC 
              || elf.e_machine != EM_386) {
      close(fd);
      return 1;
   }      

   pd->entry = elf.e_entry;
   pd->stackTop = 0xC0000000;
   pd->stack = NULL;
 
   lseek(fd, elf.e_phoff, SEEK_SET);
   for (int i = 0; i < elf.e_phnum; i++) {
      if (read(fd, &ph, sizeof(Elf32_Phdr)) != sizeof(Elf32_Phdr)) {
         close(fd);
         return 1;
      }
      if (ph.p_type == PT_LOAD) {
         if (ph.p_vaddr < org) {
            org = ph.p_vaddr & 0xFFFFF000; //assumes p_align == 0x1000
         }
         uint end = ph.p_vaddr + ph.p_memsz;
         if (end > maxaddr) maxaddr = end;
      }
   }
   if (maxaddr) {
      pd->org = org;
      pd->programSize = maxaddr - org;
      pd->program = (uchar*) calloc(pd->programSize, 1);
      if (pd->program) {
         for (int i = 0; i < elf.e_phnum; i++) {
            lseek(fd, elf.e_phoff + i * sizeof(Elf32_Phdr), SEEK_SET);
            read(fd, &ph, sizeof(Elf32_Phdr));
            if (ph.p_type == PT_LOAD) {
               lseek(fd, ph.p_offset, SEEK_SET);
               uchar *base = pd->program + (ph.p_vaddr - org);
               uint needed = ph.p_filesz;
               uint nRead;
               while (needed && (nRead = read(fd, base, needed)) != needed) {
                  if (nRead <= 0) {
                     fprintf(stderr, "Error reading file\n");
                     return 1;
                  }
                  base += nRead;
                  needed -= nRead;
               }    
            }
         }
      }
      else {
         fprintf(stderr, "Failed to allocate program memory\n");
         close(fd);
         return 1;
      }
   }
   return 0;
}

int loadData(int fd, ProgramData *pd) {
   read(fd, &pd->programSize, 4);
   read(fd, &pd->org, 4);

   pd->stackTop = 0xC0000000;
   pd->entry = pd->org;
   pd->stack = NULL;
 
   pd->program = (uchar*) malloc(pd->programSize);
   if (pd->program) {
      uchar *base = pd->program;
      uint needed = pd->programSize;
      uint nRead;
      while (needed && (nRead = read(fd, base, needed)) != needed) {
         if (nRead <= 0) {
            fprintf(stderr, "Error reading file\n");
            return 1;
         }
         base += nRead;
         needed -= nRead;
      }
   }
   else {
      fprintf(stderr, "Failed to allocate program memory\n");
      close(fd);
      return 1;
   }
   return 0;
}

void unwrapLayerOne(ProgramData *pd) {
   while (cpu.eip >= pd->entry) executeInstruction();
}

uint sanityCheck(ProgramData *pd) {
   uint checkval = *(uint*)checkme;
   uint index = (pd->entry - pd->org) / 4;  //start at the entry point
   uint *program = (uint*) pd->program;
   uint end = pd->programSize / 4;
   while (index < end) {
      if (program[index] == checkval) {
         if (!memcmp(&program[index], checkme, 8)) {
            return index * 4 + pd->org;
         }
      }
      index++;
   }
   return 0;
}

void dumpImage(ProgramData *pd, char *file) {
   int fd = open(file, O_CREAT | O_TRUNC | O_WRONLY, 0644);
   uint total = pd->programSize;
   uint nBytes;
   uchar *ptr = pd->program;
   while (total && ((nBytes = write(fd, ptr, total)) != total)) {
      total -= nBytes;
      ptr += nBytes;
      if (nBytes == 0xFFFFFFFF) break;
   }
   close(fd);
}

void recoverKeys(ProgramData *pd, uint v_funcBase, 
                 uint keyFuncOffsets[], uint keys[][4]) {
   for (int i = 0; i < NUMKEYS; i++) {
      cpu.eip = v_funcBase + keyFuncOffsets[i];
      esp = pd->stackTop - 0x10;
      push(esp, SIZE_DWORD); //function parameter
      push(0, SIZE_DWORD);       //return address
      while (cpu.eip) {
         executeInstruction();
      }
      pop(SIZE_DWORD); //clean off arg
      for (int j = 0; j < 4; j++) {
         keys[i][j] = pop(SIZE_DWORD);
      }
   }
}

void decryptBlocks(uchar *patchDataStart, uint keys[][4]) {
   uint nPatches = *(uint*)patchDataStart;
   uchar *blockDataStart = patchDataStart + 4 + nPatches * sizeof(Patch);
   uint nBlocks = *(uint*)blockDataStart;
   uchar *block = blockDataStart + 4;
   for (int i = 0; i < nBlocks; i++) {
      BlockHeader *bh = (BlockHeader*) (block);
      uint keyNum = (bh->magic >> 16) & 0xF;
      TEA_block_decrypt(block + 8, bh->size - 8, keys[keyNum]);
      block += bh->size;
   }
}

List *addToList(List *head, List *newItem) {
   List *curr = head, *prev = NULL;
   while (curr && curr->block->addr < newItem->block->addr) {
      prev = curr;
      curr = curr->next;
   }      
   newItem->next = curr;
   if (prev) {
      prev->next = newItem;
      return head;
   }
   return newItem;
}

//add some fill characters. Returns number of characters written
uint pad(int fd, uint start, uint end) {
   uchar fill = 0;
   uint i = start;
   for (; i < end; i++) {
      write(fd, &fill, 1);
   }
   return i - start;
}

uint writeSegment(List *head, Elf32_Phdr *p, uint curr_filesz, int fd) {
   while (head && (head->block->addr < p->p_vaddr)) {
      head = head->next;
   }
   if (!head) return curr_filesz;
   uint lastAddr = head->block->addr;
   uint maxAddr = p->p_vaddr + p->p_filesz;
   curr_filesz += pad(fd, curr_filesz, p->p_offset);
   for (; head && head->block->addr < maxAddr; head = head->next) {
      uint len = pad(fd, lastAddr, head->block->addr);
      len += write(fd, head->block->data, head->block->filesz);
      lastAddr += len;
      curr_filesz += len;
   }
   curr_filesz += pad(fd, curr_filesz, p->p_offset + p->p_filesz);
   return curr_filesz;
}

List *blockContaining(List *head, uint addr) {
   for (List *c = head; c; c = c->next) {
      if (c->block->addr <= addr && addr < (c->block->addr + c->block->memsz)) {
         return c;
      }
   }
   return NULL;
}

void patchBlocks(Patch *patchArray, uint nPatches, List *head) {
   uint disp;
   for (int i = 0; i < nPatches; i++) {
      uint loc = patchArray[i].loc;
      List *b = blockContaining(head, loc);
      if (b) {
         uchar *pLoc = &b->block->data[loc - b->block->addr];
         switch (patchArray[i].type) {
            case PUSH_r:
               *pLoc = 0x50 + patchArray[i].op1;
               break;
            case CALL:
               *pLoc++ = 0xE8;
               *(uint*)pLoc = patchArray[i].op1 - loc - 5;
               break;
            case JUMP:  //need to handle short and near
               disp = patchArray[i].op1 - loc - 2; 
               if (disp <= 0x7F || disp >= 0xFFFFFF80) {  //can it be reached by a short?
                  *pLoc++ = 0xEB;
                  *(uchar*)pLoc = (uchar) disp;
               }
               else {
                  *pLoc++ = 0xE9;
                  *(uint*)pLoc = disp - 3;
               }                  
               break;
            case PUSH_Imm32:
               *pLoc++ = 0x68;
               *(uint*)pLoc = patchArray[i].op1;
               break;
            case JMP_NearInd:
               *pLoc++ = 0xFF;
               *pLoc++ = 0x25;
               *(uint*)pLoc = patchArray[i].op1;
               break;
            case PUSH_disp_base:
               *pLoc++ = 0xFF;
               if (patchArray[i].len == 3) {
                  *pLoc++ = 0x70 + patchArray[i].op1;
                  *pLoc = patchArray[i].op2;
               }
               else {
                  *pLoc++ = 0xB0 + patchArray[i].op1;
                  *(uint*)pLoc = patchArray[i].op2;
               }
               break;
         }
      }
   }
}

void dumpToElf(uchar *patchDataStart, int fd) {
   Elf32_Ehdr elfHeader = {
      {0x7F, 'E', 'L', 'F', ELFCLASS32, ELFDATA2LSB, EV_CURRENT},
      ET_EXEC, EM_386, EV_CURRENT, 0, sizeof(Elf32_Ehdr), 0, 0, 
      sizeof(Elf32_Ehdr), sizeof(Elf32_Phdr), 0, sizeof(Elf32_Shdr), 0, 0
   };
   uint nPatches = *(uint*)patchDataStart;
   Patch *patchArray = (Patch*) (patchDataStart + 4);
   uchar *blockDataStart = patchDataStart + 4 + nPatches * sizeof(Patch);
   uint nBlocks =  *(uint*)blockDataStart;
   uchar *block = blockDataStart + 4;

   List *head = NULL;
   Block0 *zero;
   Elf32_Phdr *phtable = NULL;
   
   for (int i = 0; i < nBlocks; i++) {
      BlockHeader *bh = (BlockHeader*) (block);
      uint size = bh->size - 8;
      switch (bh->magic) {
         case 0x0090DFD1:
            zero = (Block0*) (block + 8);
            break;
         case 0x3E419837: //don't need these to rebuild original ELF
         case 0xE0A4C5D1:
            break;
         case 0x4E52B8D9:
         case 0x87030BD5: {
            List *temp = (List*) calloc(1, sizeof(List));
            temp->block = (ProgramBlock*) (block + 8);
            if (temp->block->addr == zero->phtable) {
               phtable = (Elf32_Phdr*) temp->block->data;
            }
            head = addToList(head, temp);
            break;
         }
      }   
      block += bh->size;
   }
   
   patchBlocks(patchArray, nPatches, head);
   
   elfHeader.e_entry = zero->e_entry;
//   elfHeader.e_phoff = ;  //assuming 0x34 for the moment
   elfHeader.e_phnum = zero->phnum;
   
   uint filesz = write(fd, &elfHeader, sizeof(Elf32_Ehdr));

   //assumes PT_LOAD headers are in increasing order by p_vaddr
   for (int i = 0; i < elfHeader.e_phnum; i++) {
      if (phtable[i].p_type != PT_LOAD) continue;
      filesz = writeSegment(head, &phtable[i], filesz, fd);
   }

}

int main(int argc, char **argv) {
   if (argc < 2 || argc > 3) {
      fprintf(stderr, "usage: %s [-p] <shivafile>\n", argv[0]);
      exit(1);
   }
   if (argc == 3 && strcmp(argv[1], "-p")) {
      fprintf(stderr, "Invalid option %s\n", argv[1]);
      fprintf(stderr, "usage: %s [-p] <shivafile>\n", argv[0]);
      exit(1);
   }

   char *shivaFile = argv[argc - 1];
   bool lkmdump = argc == 3;  //is the input from the lkm dump?

   int shivafd = open(shivaFile, O_RDONLY);
   if (shivafd == -1) {
      fprintf(stderr, "Unable to open input file\n", shivaFile);
      exit(1);
   }
   char *elfFile = (char*) malloc(strlen(shivaFile) + 5);
   strcpy(elfFile, shivaFile);
   strcat(elfFile, ".elf");
   int stripfd = open(elfFile, O_CREAT | O_WRONLY | O_TRUNC, 0744);
   free(elfFile);
   if (stripfd == -1) {
      fprintf(stderr, "Unable to create/open output file\n", shivaFile);
      exit(1);
   }
   
   ProgramData pd;
   int error = lkmdump ? loadData(shivafd, &pd) : loadElf(shivafd, &pd);
   if (error) {
      fprintf(stderr, "File load failed. Is this a elf/shiva file?\n");
      exit(1);
   }
   close(shivafd);
   setProgram(&pd);

   if (!lkmdump) {
      unwrapLayerOne(&pd);
   }

   uint checkme = sanityCheck(&pd);
   if (!checkme) {
      fprintf(stderr, "Could not locate shiva data block\n");
      exit(1);
   }
   
   uint v_base = checkme + 8;
   uchar *shivaData = pd.program + v_base - pd.org;

   uint keys[NUMKEYS][4];
   uint *keyFuncOffsets = (uint*) (shivaData + KEYMAT_SIZE);
   uint v_funcBase = v_base + KEY_DATA_SIZE;
   recoverKeys(&pd, v_funcBase, keyFuncOffsets, keys);

   uchar *patchDataStart = shivaData + KEY_DATA_SIZE + keyFuncOffsets[NUMKEYS];
   decryptBlocks(patchDataStart, keys);
   dumpToElf(patchDataStart, stripfd);

   close(stripfd);
   return 0;
}
