/*
   Headers for x86 emulator
   Copyright (c) 2003, Chris Eagle
   
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option) 
   any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
   more details.
   
   You should have received a copy of the GNU General Public License along with 
   this program; if not, write to the Free Software Foundation, Inc., 59 Temple 
   Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef __X86DEFS_H
#define __X86DEFS_H

#ifndef __IDP__

#ifndef WIN32

#include <sys/types.h>
typedef int64_t quad;
typedef u_int64_t uquad;

#else

typedef __int64 quad;
typedef unsigned __int64 uquad;

#endif

typedef unsigned int   uint;
typedef unsigned short ushort;
typedef unsigned char  uchar;

#else   //#ifdef __IDP__
#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>

typedef __int64 quad;
typedef unsigned __int64 uquad;
#endif

typedef uchar  byte;
typedef ushort word;
typedef uint   dword;
typedef uquad  qword;

#define CARRY 0x1
#define PARITY 0x4
#define AUX_CARRY 0x10
#define ZERO  0x40
#define SIGN 0x80
#define TRAP 0x100
#define INTERRUPT 0x200
#define DIRECTION 0x400
#define OVERFLOW 0x800

#define CF CARRY
#define PF PARITY
#define AF AUX_CARRY
#define ZF ZERO
#define SF SIGN
#define TF TRAP
#define IF INTERRUPT
#define DF DIRECTION
#define OF OVERFLOW

#define D (cpu.eflags & DF)

#define SET(x) (cpu.eflags |= (x))
#define CLEAR(x) (cpu.eflags &= (~x))

#define O (cpu.eflags & OF)
#define NO (!(cpu.eflags & OF))

#define B (cpu.eflags & CF)
#define C B
#define NAE B
#define NB (!(cpu.eflags & CF))
#define AE NB
#define NC NB

#define E (cpu.eflags & ZF)
#define Z E
#define NE (!(cpu.eflags & ZF))
#define NZ NE

#define BE (cpu.eflags & (ZF | CF))
#define NA BE
#define NBE (!(cpu.eflags & (ZF | CF)))
#define A NBE

#define S (cpu.eflags & SF)
#define NS (!(cpu.eflags & SF))

#define P (cpu.eflags & PF)
#define PE P
#define NP (!(cpu.eflags & PF))
#define PO NP

#define L (((cpu.eflags & (SF | OF)) == SF) || \
          ((cpu.eflags & (SF | OF)) == OF))
#define NGE L
#define NL (((cpu.eflags & (SF | OF)) == 0) || \
           ((cpu.eflags & (SF | OF)) == (SF | OF)))
#define GE NL

#define LE (((cpu.eflags & (SF | OF)) == SF) || \
           ((cpu.eflags & (SF | OF)) == OF)  || Z)
#define NG LE
#define NLE ((((cpu.eflags & (SF | OF)) == 0) || \
            ((cpu.eflags & (SF | OF)) == (SF | OF))) && NZ)
#define G NLE

#define H_MASK 0x0000FF00

#define EAX 0
#define ECX 1
#define EDX 2
#define EBX 3
#define ESP 4
#define EBP 5
#define ESI 6
#define EDI 7

#define eax (cpu.general[EAX])
#define ecx (cpu.general[ECX])
#define edx (cpu.general[EDX])
#define ebx (cpu.general[EBX])
#define esp (cpu.general[ESP])
#define ebp (cpu.general[EBP])
#define esi (cpu.general[ESI])
#define edi (cpu.general[EDI])

#define MOD_0 0
#define MOD_1 0x40
#define MOD_2 0x80
#define MOD_3 0xC0

#define RM(x)     ((x) & 0x07)
#define MOD(x)    ((x) & 0xC0)
#define REG(x) (((x) >> 3) & 0x07)
#define HAS_SIB(x) (RM(x) == 4)

#define SCALE(x) (1 << ((x) >> 6))
#define INDEX(x) (((x) >> 3) & 0x07)
#define BASE(x)  ((x) & 0x07)

typedef struct _regs {
   dword general[8];
   dword eip;
   dword eflags;
} regs;

extern regs cpu;

#define PREFIX_LOCK  0x01
#define PREFIX_REPNE 0x02
#define PREFIX_REP   0x04
#define PREFIX_CS    0x08
#define PREFIX_SS    0x10
#define PREFIX_DS    0x20
#define PREFIX_ES    0x40
#define PREFIX_FS    0x80
#define PREFIX_GS    0x100
#define PREFIX_SIZE    0x200
#define PREFIX_ADDR    0x400
#define PREFIX_SIMD    0x800

//operand types
#define TYPE_REG  1
#define TYPE_IMM  2
#define TYPE_MEM  4

//operand sizes
#define SIZE_BYTE 1
#define SIZE_WORD 2
#define SIZE_DWORD 4

//masks to clear out bytes appropriate to the sizes above
extern dword SIZE_MASKS[5];

//masks to clear out bytes appropriate to the sizes above
extern dword SIGN_BITS[5];

//masks to clear out bytes appropriate to the sizes above
extern qword CARRY_BITS[5];

extern byte BITS[5];

//struct to describe an instruction being decoded
typedef struct _inst {
   dword source;  //source address or register (depends on srcType)
   dword dest;    //dest address or register (depends on srcType)
   dword opsize;  //operand size for this instruction
   dword prefix;  //any prefix flags
   byte opcode;   //opcode, first or second byte (if first == 0x0F)
   byte srcType;  //type identifier for source operand
   byte destType; //type identifier for dest operand
} inst;

//linked list of dirty memory locations to assist in display updates
struct DirtyList {
   DirtyList(dword address, DirtyList *nxt) : addr(address), next(nxt) {};
   dword addr;
   DirtyList *next;
};

extern DirtyList *dirtyHead;
void freeDirtyList(void);

struct ProgramData {
   uint org;
   uint programSize;
   uchar *program;
   uint entry;
   uint stackSize;
   uint stackTop;
   uchar *stack;
};

void setProgram(ProgramData *pd);

void initStack(unsigned int size);
void growStack(unsigned int more);
void shrinkStack(unsigned int less);

void push(dword val, byte size);
dword pop(byte size);

int executeInstruction();

#endif
